import base64


def compose_kubeconfig_for_serviceaccount(
    server: str, ca: str, namespace: str, token: str
):
    return f"""apiVersion: v1
kind: Config
clusters:
  - name: default-cluster
    cluster:
      certificate-authority-data: {base64.b64encode(ca.encode("utf-8")).decode("utf-8")}
      server: {server}
contexts:
  - name: default-context
    context:
      cluster: default-cluster
      namespace: {namespace}
      user: default-user
current-context: default-context
users:
  - name: default-user
    user:
      token: {token}
    """



from .connect import *  # noqa
from .create import *  # noqa
from .create_shelf import *  # noqa
from .delete import *  # noqa
from .delete_shelf import *  # noqa
from .list import *  # noqa
from .list_shelves import *  # noqa
from .read import *  # noqa
from .heartbeat import *  # noqa
from .read_shelf import *  # noqa



import binascii
import logging
from datetime import datetime
from time import sleep

import kubernetes as k8s

from dataclasses import dataclass, field, fields
from enum import Enum
from typing import Dict, Optional, Any, Union, Type

from beiboot.configuration import (
    default_configuration,
    ClientConfiguration,
    __VERSION__,
)


logger = logging.getLogger(__name__)


class StateAndEventsMixin:
    # this mixin is used by Beiboot and Shelf
    # following class variables need to be set by class that uses this mixin
    StateClass: Optional[Union[Type["BeibootState"], Type["ShelfState"]]] = None
    # the uid from Kubernetes for this object
    uid: str
    _data: Optional[Dict[str, Any]] = None
    _config: ClientConfiguration  # check for shelf

    @property
    def state(self):
        self.fetch_object()
        return self.StateClass(self._data.get("state"))

    def fetch_object(self):
        raise NotImplementedError

    def wait_for_state(
        self, awaited_state: Union["BeibootState", "ShelfState"], timeout: int = 60
    ):
        """
        > Wait for the state of the resource to be the awaited state, or raise an error if the timeout is reached

        :param awaited_state: The state we're waiting for
        :type awaited_state: State enum class
        :param timeout: The maximum time to wait for the state to change, defaults to 60
        :type timeout: int (optional)
        :return: the state of the resource.
        """
        _i = 0
        while _i < timeout:
            if self.state == awaited_state:
                logger.info(
                    f"Done waiting for state {awaited_state.value} (is: {self.state.value}, {_i}s/{timeout}s) "
                )
                return
            else:
                logger.info(
                    f"Waiting for state {awaited_state.value} (is: {self.state.value}, {_i}s/{timeout}s) "
                )
                sleep(1)
                _i = _i + 1
        if self.state != awaited_state:
            raise RuntimeError(f"Waiting for state {awaited_state.value} failed")

    @property
    def events_by_timestamp(self) -> dict[datetime, Any]:
        """
        This function returns a dictionary of events related to this resource, sorted by timestamp
        :return: A dictionary of events sorted by timestamp.
        """
        try:
            events = self._config.K8S_CORE_API.list_namespaced_event(
                namespace=self._config.NAMESPACE
            )
            related_events = list(
                filter(lambda et: et.involved_object.uid == self.uid, events.items)
            )
        except k8s.client.ApiException as e:  # type: ignore
            raise RuntimeError(str(e)) from None
        result = {}
        for revent in related_events:
            result[revent.event_time] = {
                "reason": revent.reason,
                "reporter": revent.reporting_component,
                "message": revent.message,
            }
        return dict(sorted(result.items()))


class BeibootProvider(Enum):
    K3S = "k3s"


@dataclass
class GefyraParams:
    # deploy Gefyra components to this Beiboot
    enabled: bool = field(default_factory=lambda: True)
    # endpoint written to the kubeconfig for Gefyra
    endpoint: Optional[str] = field(default_factory=lambda: None)


@dataclass
class TunnelParams:
    # deploy Tunnel components to this Beiboot
    enabled: bool = field(default_factory=lambda: True)
    # endpoint for tunnel connection
    endpoint: Optional[str] = field(default_factory=lambda: None)


@dataclass
class BeibootParameters:
    k8sVersion: Optional[str] = field(default_factory=lambda: None)
    # the ports mapped to local host
    ports: Optional[list[str]] = field(default_factory=lambda: None)
    # amount of nodes for this cluster
    nodes: Optional[int] = field(default_factory=lambda: None)
    # the max lifetime for this cluster (timedelta e.g. '2h', '2h30m')
    maxLifetime: Optional[str] = field(default_factory=lambda: None)
    # the max time for this cluster before getting removed when no client is connected
    maxSessionTimeout: Optional[str] = field(default_factory=lambda: None)
    # the max time waiting for this cluster to become 'READY'
    clusterReadyTimeout: Optional[int] = field(
        default_factory=lambda: default_configuration.CLUSTER_CREATION_TIMEOUT
    )
    # the K8s resources (requests, limits) for the server and node pods
    serverResources: Optional[dict[str, dict[str, str]]] = field(
        default_factory=lambda: None
    )
    nodeResources: Optional[dict[str, dict[str, str]]] = field(
        default_factory=lambda: None
    )
    # storage requests for server and node pods
    serverStorageRequests: Optional[str] = field(default_factory=lambda: None)
    nodeStorageRequests: Optional[str] = field(default_factory=lambda: None)
    # Gefyra component requests
    gefyra: GefyraParams = field(default_factory=lambda: GefyraParams())
    # Tunnel connection params
    tunnel: Optional[TunnelParams] = field(default_factory=lambda: TunnelParams())

    @classmethod
    def from_raw(cls, data: dict):
        params = cls()
        params.ports = data.get("ports")
        params.nodes = data.get("nodes")
        params.maxLifetime = data.get("maxLifetime")
        params.maxSessionTimeout = data.get("maxSessionTimeout")
        params.clusterReadyTimeout = data.get("clusterReadyTimeout")
        params.serverStorageRequests = data.get("serverStorageRequests")
        params.nodeStorageRequests = data.get("nodeStorageRequests")
        params.nodeResources = data.get("nodeResources")
        params.serverResources = data.get("serverResources")
        params.gefyra = data.get("gefyra")  # type: ignore
        params.tunnel = data.get("tunnel")  # type: ignore
        return params

    def as_dict(self) -> dict[str, Any]:
        data = {}
        for _field in fields(self):
            if _v := getattr(self, _field.name):
                if type(_v) == GefyraParams:
                    data["gefyra"] = {"enabled": _v.enabled, "endpoint": _v.endpoint}
                elif type(_v) == TunnelParams:
                    data["tunnel"] = {"enabled": _v.enabled, "endpoint": _v.endpoint}
                else:
                    data[_field.name] = _v
        if (data.get("tunnel") and data["tunnel"].get("endpoint")) and (
            data.get("gefyra") and data["gefyra"].get("endpoint") is None
        ):
            # if there is a special endpoint for tunnel and not for Gefyra, set it for Gefyra, too
            data["gefyra"]["endpoint"] = data["tunnel"].get("endpoint")
        return data


@dataclass
class BeibootRequest:
    # a unique name for this Beiboot cluster
    name: str
    # the K8s provider running the cluster
    provider: BeibootProvider = field(default_factory=lambda: BeibootProvider.K3S)
    parameters: BeibootParameters = field(default_factory=lambda: BeibootParameters())
    labels: Dict[str, str] = field(default_factory=lambda: {})
    from_shelf: str = field(default_factory=lambda: "")


class BeibootState(Enum):
    REQUESTED = "REQUESTED"
    PREPARING = "PREPARING"
    RESTORING = "RESTORING"
    CREATING = "CREATING"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    READY = "READY"
    TERMINATING = "TERMINATING"
    ERROR = "ERROR"


class Beiboot(StateAndEventsMixin):
    # the final name of this cluster
    name: str
    # the namespace this cluster runs in the host cluster
    namespace: str
    # the beiboot.getdeck.dev object namespace
    object_namespace: str
    # the labels of this Beiboot object
    labels: Dict[str, str]
    # the timestamp when this cluster gets removed
    sunset: Optional[str] = None
    # the timestamp when any client reported its last heartbeat
    last_client_contact: Optional[str] = None
    # all state transitions
    transitions: Optional[dict[str, str]]
    # the cluster parameters
    parameters: BeibootParameters
    provider: str

    # class variables from StateMixin (_data is set in _init_data(...))
    StateClass = BeibootState

    def __init__(
        self, beiboot: dict, config: ClientConfiguration = default_configuration
    ):
        self.name = beiboot["metadata"]["name"]
        self.uid = beiboot["metadata"]["uid"]
        self.labels = beiboot["metadata"].get("labels")
        self.object_namespace = beiboot["metadata"]["namespace"]
        self.provider = beiboot["provider"]
        self._init_data(beiboot)
        self._config = config

    def _init_data(self, _object: dict[str, Any]):
        self._data = _object
        self.namespace = str(self._data.get("beibootNamespace"))
        self.sunset = self._data.get("sunset")
        self.last_client_contact = self._data.get("lastClientContact")
        self.transitions = self._data.get("stateTransitions")
        self.parameters = BeibootParameters.from_raw(self._data["parameters"])

    def fetch_object(self):
        logger.debug(f"Fetching object Beiboot {self.name}")
        try:
            bbt = self._config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
                group="getdeck.dev",
                version="v1",
                namespace=self._config.NAMESPACE,
                plural="beiboots",
                name=self.name,
            )
        except k8s.client.exceptions.ApiException as e:  # type: ignore
            if e.status == 404:
                raise RuntimeError(
                    f"The Beiboot '{self.name}' does not exist anymore"
                ) from None
            else:
                raise RuntimeError(str(e)) from None
        self._init_data(bbt)  # type: ignore

    @property
    def kubeconfig(self) -> Optional[str]:
        """
        It returns the kubeconfig of the Beiboot.
        :return: The kubeconfig object is being returned.
        """
        from beiboot.utils import decode_kubeconfig

        if self.state != BeibootState.READY:
            logger.warning("This Beiboot is not in READY state")
        if self._data and "kubeconfig" in self._data:
            kubeconfig_object: Dict[str, str] = self._data["kubeconfig"]  # noqa
        else:
            return None
        try:
            kubeconfig = decode_kubeconfig(kubeconfig_object)
        except (KeyError, binascii.Error) as e:
            raise RuntimeError(f"Can not fetch kubeconfig: {e}") from None
        else:
            return kubeconfig

    @property
    def mtls_files(self) -> Optional[dict[str, str]]:
        """
        It returns the mTLS files for the Beiboot if it's in the READY state
        :return: The mtls_files property returns a dictionary of the mTLS files.
        """
        from beiboot.utils import decode_b64_dict

        if self.state != BeibootState.READY:
            logger.warning("This Beiboot is not in READY state")
        if tunnel := self.tunnel:
            if ghostunnel := tunnel.get("ghostunnel"):
                try:
                    return decode_b64_dict(ghostunnel["mtls"])
                except binascii.Error as e:
                    raise RuntimeError(
                        f"There was an error decoding mTLS data: {e}"
                    ) from None
        return None

    @property
    def serviceaccount_tokens(self) -> Optional[dict[str, Union[str, Any]]]:
        """
        It returns a dictionary holding the service account token for the Beiboot
        :return: A dictionary holding the service account tokens.
        """
        from beiboot.utils import decode_b64_dict

        if self.state != BeibootState.READY:
            logger.warning("This Beiboot is not in READY state")
        if tunnel := self.tunnel:
            if sa_token := tunnel.get("serviceaccount"):
                try:
                    return decode_b64_dict(sa_token)
                except binascii.Error as e:
                    raise RuntimeError(
                        f"There was an error decoding the serviceaccount token: {e}"
                    ) from None
        return None

    @property
    def tunnel(self) -> Optional[dict[str, Any]]:
        if self._data and "tunnel" in self._data:
            return self._data["tunnel"]
        return None


@dataclass
class InstallOptions:
    namespace: str = field(
        default_factory=lambda: "getdeck",
        metadata=dict(
            help="The namespace to install Beiboot into (default: getdeck)", short="ns"
        ),
    )
    version: str = field(
        default_factory=lambda: __VERSION__,
        metadata=dict(
            help="Set the Operator version; components are created according to this beibootctl version"
        ),
    )
    storage_class: str = field(
        default_factory=lambda: "standard",
        metadata=dict(
            help="Set the Kubernetes StorageClass of PVCs created for Beiboot clusters (default: standard)"
        ),
    )
    shelf_storage_class: str = field(
        default_factory=lambda: "standard",
        metadata=dict(
            help="Set the Kubernetes VolumeSnapshotClass of VolumeSnapshots created for Beiboot shelf clusters (default: standard)"  # noqa: E501
        ),
    )
    node_storage_request: str = field(
        default_factory=lambda: "1Gi",
        metadata=dict(
            help="The default storage request for Beiboot nodes (default: 1Gi)"
        ),
    )
    server_storage_request: str = field(
        default_factory=lambda: "1Gi",
        metadata=dict(
            help="The default storage request for Beiboot servers (default: 1Gi)"
        ),
    )
    nodes: str = field(
        default_factory=lambda: "1",
        metadata=dict(help="The default nodes count per Beiboot (default: 1)"),
    )
    max_session_timeout: str = field(
        default_factory=lambda: "null",
        metadata=dict(
            help="The default maximum session timeout for a Beiboot cluster before it will be deleted (default: null)"
        ),
    )
    max_lifetime: str = field(
        default_factory=lambda: "null",
        metadata=dict(
            help="The default maximum lifetime for a Beiboot cluster before it will be deleted (default: null)"
        ),
    )
    namespace_prefix: str = field(
        default_factory=lambda: "getdeck-bbt",
        metadata=dict(
            help="The namespace prefix for Beiboot clusters within the host cluster (default: getdeck-bbt)",
        ),
    )
    server_requests_cpu: str = field(
        default_factory=lambda: "1",
        metadata=dict(
            help="The default CPU request for each Beiboot server pod (default: 1)",
        ),
    )
    server_requests_memory: str = field(
        default_factory=lambda: "1Gi",
        metadata=dict(
            help="The default memory request for each Beiboot server pod (default: 1Gi)",
        ),
    )
    node_requests_cpu: str = field(
        default_factory=lambda: "1",
        metadata=dict(
            help="The default CPU request for each Beiboot node pod (default: 1)",
        ),
    )
    node_requests_memory: str = field(
        default_factory=lambda: "1Gi",
        metadata=dict(
            help="The default memory request for each Beiboot node pod (default: 1Gi)",
        ),
    )


@dataclass
class VolumeSnapshotContentStatus(Enum):
    NOT_READY_TO_USE = "NOT_READY_TO_USE"
    READY_TO_USE = "READY_TO_USE"


@dataclass
class VolumeSnapshotContent:
    name: str
    snapshot_handle: str
    node: str
    status: VolumeSnapshotContentStatus = field(
        default_factory=lambda: VolumeSnapshotContentStatus.NOT_READY_TO_USE
    )


@dataclass
class ShelfRequest:
    # a unique name for this Shelf
    name: str
    cluster_name: str
    volume_snapshot_class: str = ""
    volume_snapshot_contents: list[VolumeSnapshotContent] = field(
        default_factory=lambda: []
    )
    labels: Dict[str, str] = field(default_factory=lambda: {})


class ShelfState(Enum):
    REQUESTED = "REQUESTED"
    CREATING = "CREATING"
    PREPARING = "PREPARING"
    PENDING = "PENDING"
    READY = "READY"
    TERMINATING = "TERMINATING"
    ERROR = "ERROR"


class Shelf(StateAndEventsMixin):
    # the final name of this cluster
    name: str
    # the namespace this shelf is located in the host cluster (currently it's always "getdeck")
    namespace: str
    # list of VolumeSnapshotContents that are contained in this shelf
    volume_snapshot_contents: Optional[list[VolumeSnapshotContent]]
    # the labels of this Beiboot object
    labels: Dict[str, str]
    # all state transitions
    transitions: Optional[dict[str, str]]

    # class variables from StateMixin (_data is set in _init_data(...))
    StateClass = ShelfState

    def __init__(
        self, shelf: dict, config: ClientConfiguration = default_configuration
    ):
        self.name = shelf["metadata"]["name"]
        self.uid = shelf["metadata"]["uid"]
        self.labels = shelf["metadata"].get("labels")
        self._init_data(shelf)
        self._config = config

    def _init_data(self, _object: dict[str, Any]):
        self._data = _object
        self.transitions = self._data.get("stateTransitions")
        self.volume_snapshot_contents = self._data.get("volumeSnapshotContents")

    def fetch_object(self):
        logger.debug(f"Fetching object Shelf {self.name}")
        try:
            shelf = self._config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
                group="beiboots.getdeck.dev",
                version="v1",
                namespace=self._config.NAMESPACE,
                plural="shelves",
                name=self.name,
            )
        except k8s.client.exceptions.ApiException as e:  # type: ignore
            if e.status == 404:
                raise RuntimeError(
                    f"The Shelf '{self.name}' does not exist anymore"
                ) from None
            else:
                raise RuntimeError(str(e)) from None
        self._init_data(shelf)  # type: ignore



import kubernetes as k8s

from beiboot.api import stopwatch
from beiboot.configuration import ClientConfiguration, default_configuration
from beiboot.types import Beiboot


@stopwatch
def read(name: str, config: ClientConfiguration = default_configuration) -> Beiboot:
    """
    Reads a Beiboot from the Kubernetes API.

    :param name: The name of the Beiboot to read.
    :param config: The configuration to use.
    :return: The Beiboot.
    :raises RuntimeError: If the Beiboot does not exist.
    """
    try:
        bbt = config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
            group="getdeck.dev",
            version="v1",
            namespace=config.NAMESPACE,
            plural="beiboots",
            name=name,
        )
    except k8s.client.ApiException as e:  # type: ignore
        if e.status == 404:
            raise RuntimeError(f"The Beiboot {name} does not exist")
        raise RuntimeError(str(e)) from None
    return Beiboot(bbt)  # type: ignore



from beiboot.types import InstallOptions


def data(params: InstallOptions) -> list[dict]:
    return [
        {
            "apiVersion": "v1",
            "kind": "ConfigMap",
            "metadata": {
                "name": "beiboot-config",
                "namespace": params.namespace,
            },
            "data": {
                "clusterReadyTimeout": "180",
                "gefyra": '{"enabled": true, "endpoint": null}',
                "k8sVersion": "null",
                "maxLifetime": params.max_lifetime,
                "maxSessionTimeout": params.max_session_timeout,
                "namespacePrefix": params.namespace_prefix,
                "nodeLabels": '{"app": "beiboot", "beiboot.getdeck.dev/is-node": "true"}',
                "nodeResources": f'{{"requests": {{"cpu": "{params.node_requests_cpu}", "memory": "{params.node_requests_memory}"}}, "limits": {{}}}}',  # noqa
                "nodeStorageRequests": params.node_storage_request,
                "nodes": params.nodes,
                "ports": "null",
                "serverLabels": '{"app": "beiboot", "beiboot.getdeck.dev/is-node": "true", "beiboot.getdeck.dev/is-server": "true"}',  # noqa
                "serverResources": f'{{"requests": {{"cpu": "{params.server_requests_cpu}", "memory": "{params.server_requests_memory}"}}, "limits": {{}}}}',  # noqa
                "serverStartupTimeout": "60",
                "serverStorageRequests": params.server_storage_request,
                "tunnel": '{"enabled": true, "endpoint": null}',
                "storageClass": params.storage_class,
                "shelfStorageClass": params.shelf_storage_class,
            },
        }
    ]



import logging

import kubernetes as k8s

from beiboot.api import stopwatch
from beiboot.configuration import ClientConfiguration, default_configuration
from beiboot.types import Beiboot


logger = logging.getLogger(__name__)


@stopwatch
def delete(bbt: Beiboot, config: ClientConfiguration = default_configuration) -> None:
    """
    Mark a Beiboot for deletion

    :param bbt: The Beiboot to be marked for deletion
    :type bbt: Beiboot
    """
    _delete_bbt(bbt.name, config)


@stopwatch
def delete_by_name(
    name: str, config: ClientConfiguration = default_configuration
) -> None:
    """
    Mark a Beiboot for deletion

    :param name: The Beiboot name to be marked for deletion
    :type name: str
    """
    _delete_bbt(name, config)


def _delete_bbt(cluster_name: str, config: ClientConfiguration):
    logger.debug(f"Now removing Beiboot {cluster_name}")
    try:
        config.K8S_CUSTOM_OBJECT_API.delete_namespaced_custom_object(
            namespace=config.NAMESPACE,
            name=cluster_name,
            group="getdeck.dev",
            plural="beiboots",
            version="v1",
        )
        logger.debug(f"Successfully deleted Beiboot {cluster_name}")
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            #  Getdeck Beiboot probably not available
            raise RuntimeWarning(f"Beiboot {cluster_name} does not exist")
        else:
            raise RuntimeError(
                f"Error deleting Beiboot object: {e.reason} ({e.status})"
            )



from typing import Dict
import kubernetes as k8s

from beiboot.api import stopwatch
from beiboot.configuration import ClientConfiguration, default_configuration
from beiboot.types import Shelf


@stopwatch
def read_all_shelves(
    labels: Dict[str, str] = {}, config: ClientConfiguration = default_configuration
) -> list[Shelf]:
    """
    Reads all Shelves.

    :param labels: A dictionary of labels to filter the Shelves by.
    :param config: The configuration to use.
    :return: A list of Shelves.
    """
    result = []
    if labels:
        _labels = ",".join([f"{label}={value}" for label, value in labels.items()])
    else:
        _labels = ""
    try:
        shelves = config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object(
            group="beiboots.getdeck.dev",
            version="v1",
            namespace=config.NAMESPACE,
            plural="shelves",
            label_selector=_labels,
        )
    except k8s.client.ApiException as e:  # type: ignore
        if e.status == 404:
            raise RuntimeError(
                "This cluster does probably not support Getdeck Beiboot, or is not ready."
            ) from None
        else:
            raise RuntimeError(str(e)) from None
    else:
        for shelf in shelves["items"]:
            result.append(Shelf(shelf))
    return result



import os.path
import pathlib
import shutil
from abc import ABC, abstractmethod
from typing import Optional, List, Iterable

from beiboot.configuration import ClientConfiguration
from beiboot.connection.utils import compose_kubeconfig_for_serviceaccount
from beiboot.types import Beiboot
from beiboot.utils import get_kubeconfig_location, get_beiboot_config_location


class AbstractConnector(ABC):
    connector_type = ""
    beiboot: Beiboot
    additional_ports: Optional[List[str]]

    def __init__(
        self,
        configuration: ClientConfiguration,
    ) -> None:
        self.configuration = configuration

    @abstractmethod
    def establish(
        self,
        beiboot: Beiboot,
        additional_ports: Optional[List[str]],
        host: Optional[str],
    ) -> None:
        """
        Establishes a connection to a Beiboot

        :param beiboot: The Beiboot object that is being used to establish the connection
        :type beiboot: Beiboot
        :param additional_ports: A list of ports that should be forwarded to the host
        :type additional_ports: Optional[List[str]]
        :param host: The hostname or IP address of the host to connect to
        :type host: Optional[str]
        """
        raise NotImplementedError

    @abstractmethod
    def terminate(self, name: str) -> None:
        """
        Terminate the connection, clean up

        :param name: The name of the process to terminate
        :type name: str
        """
        raise NotImplementedError

    def _save_items(
        self, beiboot_name: str, items: Iterable, prefix: str = "."
    ) -> dict[str, str]:
        location = os.path.join(
            get_beiboot_config_location(self.configuration, beiboot_name), prefix
        )
        pathlib.Path(location).mkdir(parents=True, exist_ok=True)
        file_locations = {}
        for fname, content in items:
            _p = str(pathlib.Path(location).joinpath(fname))
            file_locations[fname] = _p
            with open(_p, "w") as f:
                f.write(content)
        return file_locations

    def save_mtls_files(self, beiboot: Beiboot) -> dict[str, str]:
        _mtls_files = beiboot.mtls_files
        if not _mtls_files:
            raise RuntimeError("Cannot write mTLS files")
        return self._save_items(beiboot.name, _mtls_files.items(), "mtls")

    def save_serviceaccount_files(self, beiboot: Beiboot) -> dict[str, str]:
        _sa_tokens = beiboot.serviceaccount_tokens
        if not _sa_tokens:
            raise RuntimeError("Cannot write Service account files")
        sa_files = self._save_items(beiboot.name, _sa_tokens.items(), "serviceaccount")
        location = get_beiboot_config_location(self.configuration, beiboot.name)
        try:
            import kubernetes as k8s

            api_host = k8s.client.api_client.ApiClient().configuration.host
            sa_kubeconfig = str(pathlib.Path(location).joinpath("sa_kubeconfig.yaml"))
            with open(sa_kubeconfig, "w") as f:
                f.write(
                    compose_kubeconfig_for_serviceaccount(
                        api_host,
                        _sa_tokens["ca.crt"],
                        _sa_tokens["namespace"],
                        _sa_tokens["token"],
                    )
                )
            sa_files.update({"sa_kubeconfig.yaml": sa_kubeconfig})
        except Exception as e:
            print(e)
        return sa_files

    def save_kubeconfig_to_file(self, beiboot: Beiboot) -> str:
        _kubeconfig = beiboot.kubeconfig
        if not _kubeconfig:
            raise RuntimeError("Cannot write kubeconfig")
        location = get_kubeconfig_location(self.configuration, beiboot.name)
        with open(location, "w") as yaml_file:
            yaml_file.write(_kubeconfig)
        return location

    def delete_kubeconfig_file(self, beiboot_name: str):
        location = get_kubeconfig_location(self.configuration, beiboot_name)
        pathlib.Path(location).unlink(missing_ok=True)

    def delete_beiboot_config_directory(self, beiboot_name):
        location = get_beiboot_config_location(self.configuration, beiboot_name)
        shutil.rmtree(location, ignore_errors=True)



import getpass
import logging
import socket
from typing import Optional, List

import docker

from beiboot.configuration import ClientConfiguration
from beiboot.connection.abstract import AbstractConnector
from beiboot.connection.types import ConnectorType
from beiboot.types import Beiboot
from beiboot.utils import _list_containers_by_prefix

logger = logging.getLogger(__name__)


class GhostunnelDockerBuilder:
    def __init__(self):
        self._instances = {}

    def __call__(
        self,
        configuration: ClientConfiguration,
        **_ignored,
    ):
        instance = GhostunnelDocker(
            configuration=configuration,
        )
        return instance


class GhostunnelDocker(AbstractConnector):

    connector_type = ConnectorType.GHOSTUNNEL_DOCKER.value
    IMAGE = "ghostunnel/ghostunnel:v1.7.1"
    HEARTBEAT_IMAGE = "quay.io/getdeck/tooler:latest"
    CONTAINER_PREFIX = "getdeck-beiboot-{name}"
    _DOCKER_NETWORK_NAME = None
    HEARTBEAT_CMD = 'watch -n30 "VAR=\'{{\\"data\\":{{\\"{client}\\": \\"\'$(date +\\"%Y-%m-%dT%H:%M:%S\\")\'\\"}}}}\';  kubectl --kubeconfig /kubernetes/sa_kubeconfig.yaml patch configmap beiboot-clients -n {namespace} --type merge --patch=\\"$VAR\\""'  # noqa
    CLIENT = f"{socket.gethostname()}-{getpass.getuser()}"

    def __init__(
        self,
        configuration: ClientConfiguration,
    ):
        super(GhostunnelDocker, self).__init__(configuration)
        self.HEARTBEAT_IMAGE = configuration.TOOLER_IMAGE

    def set_docker_network(self, network_name):
        self._DOCKER_NETWORK_NAME = network_name

    def establish(
        self,
        beiboot: Beiboot,
        additional_ports: Optional[List[str]],
        host: Optional[str],
    ) -> None:
        if not additional_ports:
            additional_ports = []
        if beiboot.tunnel is None:
            raise RuntimeError(
                "Connection data is not available, unable to establish connection"
            )

        ghostunnel = beiboot.tunnel["ghostunnel"]
        remote_ports = ghostunnel.get(
            "ports"
        )  # "endpoint": -> address; "target": cluster port

        def _nodeport_for_target(target: str) -> Optional[str]:
            for port in remote_ports:
                if str(port.get("target")) == target:
                    _host, _port = port.get("endpoint").split(":")
                    if (not bool(_host) or _host == "None") and not bool(host):
                        raise RuntimeError(
                            "Cannot connect to this Beiboot, as there is no endpoint available."
                        )
                    return f"{host or _host}:{_port}"
            return None

        additional_ports.extend(beiboot.parameters.ports or [])
        mtls_files = self.save_mtls_files(beiboot)  # {filename, path}
        serviceaccount_files = self.save_serviceaccount_files(
            beiboot
        )  # {filename, path}
        container_prefixes = self.CONTAINER_PREFIX.format(name=beiboot.name)

        for call_port in additional_ports:
            local_port, cluster_port = call_port.split(":")
            _endpoint = _nodeport_for_target(cluster_port)
            if not _endpoint:
                logger.warning(f"No endpoint for cluster port {cluster_port}")
                continue
            _cmd = (
                f"client --listen 0.0.0.0:{local_port} --unsafe-listen "
                f"--target {_endpoint} --cert /crt/client.crt --key /crt/client.key --cacert /crt/ca.crt"
            )
            try:
                container = self.configuration.DOCKER.containers.run(  # noqa
                    image=self.IMAGE,
                    name=f"{container_prefixes}-{cluster_port}",
                    command=_cmd,
                    restart_policy={"Name": "unless-stopped"},
                    remove=False,
                    detach=True,
                    ports={f"{local_port}": int(local_port)},
                    volumes=[
                        f"{path}:/crt/{file}" for file, path in mtls_files.items()
                    ],
                    network=self._DOCKER_NETWORK_NAME or None,
                )
            except docker.errors.APIError as e:
                logger.critical(e)
                raise RuntimeError(
                    f"Could not run ghostunnel container due to the following error: {e}"
                ) from None

        try:

            container = self.configuration.DOCKER.containers.run(  # noqa
                image=self.HEARTBEAT_IMAGE,
                name=f"{container_prefixes}-heartbeat",
                command=self.HEARTBEAT_CMD.format(
                    client=self.CLIENT, namespace=beiboot.namespace
                ),
                restart_policy={"Name": "unless-stopped"},
                remove=False,
                detach=True,
                volumes=[
                    f"{path}:/kubernetes/{file}"
                    for file, path in serviceaccount_files.items()
                ],
                network=self._DOCKER_NETWORK_NAME or None,
            )
        except docker.errors.APIError as e:
            raise RuntimeError(
                f"Could not run heartbeat container due to the following error: {e}"
            ) from None

    def terminate(self, name: str) -> None:
        try:
            containers = _list_containers_by_prefix(
                self.configuration, self.CONTAINER_PREFIX.format(name=name)
            )
            for container in containers:
                try:
                    container.kill()  # type: ignore
                except:  # noqa
                    pass
                try:
                    container.remove()  # type: ignore
                except:  # noqa
                    pass
        except docker.errors.APIError as e:
            logger.warning(str(e))



# flake8: noqa
from beiboot.types import InstallOptions


def data(params: InstallOptions) -> list[dict]:
    return [
        {
            "apiVersion": "v1",
            "kind": "Namespace",
            "metadata": {"name": params.namespace},
        }
    ]



import logging
from datetime import datetime
from typing import Optional
from beiboot.configuration import ClientConfiguration, default_configuration
import kubernetes as k8s

from beiboot.api import stopwatch
from beiboot.types import Beiboot

logger = logging.getLogger(__name__)


@stopwatch
def write_heartbeat(
    client_id: str,
    bbt: Beiboot,
    timestamp: Optional[datetime] = None,
    config: ClientConfiguration = default_configuration,
) -> datetime:
    """
    Create a client contact entry to the heartbeat configmap for a Beiboot. Returns the timestamp written to the
    configmap.

    :param client_id: The client id to write the contact entry on behalf of
    :type client_id: str
    :param bbt: The Beiboot to write a heartbeat for
    :type bbt: Beiboot
    :param timestamp: Optional timestamp to write to the configmap
    :type timestamp: datetime
    """

    if timestamp is None:
        timestamp = datetime.utcnow()

    _timestamp = timestamp.isoformat()
    configmap = k8s.client.V1ConfigMap(  # type: ignore
        api_version="v1",
        kind="ConfigMap",
        data={client_id: _timestamp},
        metadata=k8s.client.V1ObjectMeta(  # type: ignore
            name=config.CLIENT_HEARTBEAT_CONFIGMAP_NAME,
            namespace=bbt.namespace,
        ),
    )
    try:
        config.K8S_CORE_API.patch_namespaced_config_map(
            name=config.CLIENT_HEARTBEAT_CONFIGMAP_NAME,
            namespace=bbt.namespace,
            body=configmap,
        )
        logger.debug(f"Successfully heartbeat for client {client_id} to {_timestamp}")
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            raise RuntimeError(
                f"Cannot write heartbeat, the required configmap '{config.CLIENT_HEARTBEAT_CONFIGMAP_NAME}' does not exist"
            ) from None
        else:
            raise RuntimeError(f"Cannot write heartbeat: {e}") from None

    return timestamp



import logging
from typing import Optional, List

from beiboot.configuration import ClientConfiguration
from beiboot.connection.abstract import AbstractConnector
from beiboot.types import Beiboot

logger = logging.getLogger(__name__)


class DummyNoConnectBuilder:
    def __init__(self):
        self._instances = {}

    def __call__(
        self,
        configuration: ClientConfiguration,
        **_ignored,
    ):
        instance = DummyNoConnect(
            configuration=configuration,
        )
        return instance


class DummyNoConnect(AbstractConnector):
    def establish(
        self,
        beiboot: Beiboot,
        additional_ports: Optional[List[str]],
        host: Optional[str],
    ) -> None:
        _ = self.save_mtls_files(beiboot)
        _ = self.save_serviceaccount_files(beiboot)

        def _nodeport_for_target(target: str) -> Optional[str]:
            for port in remote_ports:
                if str(port.get("target")) == target:
                    _host, _port = port.get("endpoint").split(":")
                    if (not bool(_host) or _host == "None") and not bool(host):
                        return "Unavailable"
                    return f"{host or _host}:{_port}"
            return None

        ghostunnel = beiboot.tunnel["ghostunnel"]  # type: ignore
        remote_ports = ghostunnel.get("ports")
        if not additional_ports:
            additional_ports = []
        for call_port in additional_ports:
            local_port, cluster_port = call_port.split(":")
            _endpoint = _nodeport_for_target(cluster_port)
            if not _endpoint:
                logger.warning(f"No endpoint for cluster port {cluster_port}")
                continue
            else:
                logger.info(f"Forwarding port: {local_port} -> {_endpoint}")

    def terminate(self, name: str) -> None:
        self.delete_beiboot_config_directory(name)



from beiboot.configuration import ClientConfiguration
from beiboot.connection.abstract import AbstractConnector
from beiboot.connection.dummy import DummyNoConnectBuilder
from beiboot.connection.ghostunnel import (
    GhostunnelDockerBuilder,
)
from beiboot.connection.types import ConnectorType


class ConnectorFactory:
    def __init__(self):
        self._builders = {}

    def register_builder(self, connector_type: ConnectorType, builder):
        self._builders[connector_type.value] = builder

    def __create(
        self,
        connector_type: ConnectorType,
        configuration: ClientConfiguration,
        **kwargs
    ):
        builder = self._builders.get(connector_type.value)
        if not builder:
            raise ValueError(connector_type)
        return builder(configuration, **kwargs)

    def get(
        self,
        connector_type: ConnectorType,
        configuration: ClientConfiguration,
        **kwargs
    ) -> AbstractConnector:
        return self.__create(connector_type, configuration, **kwargs)


connector_factory = ConnectorFactory()
connector_factory.register_builder(
    ConnectorType.GHOSTUNNEL_DOCKER, GhostunnelDockerBuilder()
)
connector_factory.register_builder(
    ConnectorType.DUMMY_NO_CONNECT, DummyNoConnectBuilder()
)



# flake8: noqa
from beiboot.types import InstallOptions


def data(params: InstallOptions) -> list[dict]:
    return [
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {"name": "beiboot-admission", "namespace": params.namespace},
            "spec": {
                "selector": {"app": "beiboot-operator"},
                "ports": [{"protocol": "TCP", "port": 443, "targetPort": 9443}],
            },
        },
        {
            "apiVersion": "admissionregistration.k8s.io/v1",
            "kind": "ValidatingWebhookConfiguration",
            "metadata": {"name": "beiboot.getdeck.dev"},
            "webhooks": [
                {
                    "admissionReviewVersions": ["v1", "v1beta1"],
                    "clientConfig": {
                        "caBundle": "LS0tLS1CRUdJTiBDRVJUSUZJQ0FURS0tLS0tCk1JSUZDakNDQXZLZ0F3SUJBZ0lVTzJZMlJqUVdL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\nV2QxcDVRNGlCUXFoZz09Ci0tLS0tRU5EIENFUlRJRklDQVRFLS0tLS0K\n",
                        "service": {
                            "name": "beiboot-admission",
                            "namespace": params.namespace,
                            "path": "/validate-parameters",
                        },
                    },
                    "failurePolicy": "Fail",
                    "matchPolicy": "Equivalent",
                    "name": "validate-parameters.beiboot.getdeck.dev",
                    "namespaceSelector": {},
                    "objectSelector": {},
                    "rules": [
                        {
                            "apiGroups": ["getdeck.dev"],
                            "apiVersions": ["v1"],
                            "operations": ["CREATE"],
                            "resources": ["beiboots"],
                            "scope": "*",
                        }
                    ],
                    "sideEffects": "None",
                    "timeoutSeconds": 30,
                }
            ],
        },
        {
            "apiVersion": "admissionregistration.k8s.io/v1",
            "kind": "ValidatingWebhookConfiguration",
            "metadata": {"name": "shelf.beiboots.getdeck.dev"},
            "webhooks": [
                {
                    "admissionReviewVersions": ["v1", "v1beta1"],
                    "clientConfig": {
                        "caBundle": "LS0tLS1CRUdJTiBDRVJUSUZJQ0FURS0tLS0tCk1JSUZDakNDQXZLZ0F3SUJBZ0lVTzJZMlJqUVdL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\nV2QxcDVRNGlCUXFoZz09Ci0tLS0tRU5EIENFUlRJRklDQVRFLS0tLS0K\n",
                        "service": {
                            "name": "beiboot-admission",
                            "namespace": params.namespace,
                            "path": "/validate-shelf",
                        },
                    },
                    "failurePolicy": "Fail",
                    "matchPolicy": "Equivalent",
                    "name": "validate-shelf.shelf.beiboots.getdeck.dev",
                    "namespaceSelector": {},
                    "objectSelector": {},
                    "rules": [
                        {
                            "apiGroups": ["beiboots.getdeck.dev"],
                            "apiVersions": ["v1"],
                            "operations": ["CREATE"],
                            "resources": ["shelves"],
                            "scope": "*",
                        }
                    ],
                    "sideEffects": "None",
                    "timeoutSeconds": 30,
                }
            ],
        },
    ]



import json
import logging

import kubernetes as k8s

from beiboot.api.utils import stopwatch
from beiboot.configuration import default_configuration, ClientConfiguration
from beiboot.types import ShelfRequest, Shelf
from beiboot.utils import create_shelf_custom_ressource

logger = logging.getLogger(__name__)


@stopwatch
def create_shelf(
    req: ShelfRequest, config: ClientConfiguration = default_configuration
) -> Shelf:
    """
    It creates a Shelf

    :param req: ShelfRequest
    :type req: ShelfRequest
    :param config: ClientConfiguration = default_configuration
    :type config: ClientConfiguration
    :return: A Shelf object
    """
    logger.debug("creating shelf")
    obj = create_shelf_custom_ressource(req, config)
    try:
        logger.debug("Checking if Shelf object already exists")
        _ = config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
            group="beiboots.getdeck.dev",
            version="v1",
            namespace=config.NAMESPACE,
            plural="shelves",
            name=req.name,
        )
        raise RuntimeError(
            f"The requested Shelf object '{req.name}' already exists. You can list existing shelves with 'shelf list'."
        )
    except k8s.client.exceptions.ApiException:  # type: ignore
        logger.debug("Shelf object does not exist and can be created")
        pass
    try:
        shelf = config.K8S_CUSTOM_OBJECT_API.create_namespaced_custom_object(
            namespace=config.NAMESPACE,
            body=obj,
            group="beiboots.getdeck.dev",
            plural="shelves",
            version="v1",
        )
        logger.debug(f"Successfully created Shelf object: {obj['metadata']['name']}")
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            raise RuntimeError(
                "This cluster does probably not support Getdeck Beiboot, or is not ready."
            ) from None
        elif e.status == 500:
            raise RuntimeError(
                f"The requested Shelf {req.name} cannot be created: {json.loads(e.body).get('message')}"
            ) from None
        else:
            # TODO handle this case
            raise
    _shelf = Shelf(shelf)
    return _shelf



import kubernetes as k8s

from beiboot.api import stopwatch
from beiboot.configuration import ClientConfiguration, default_configuration
from beiboot.types import Shelf


@stopwatch
def read_shelf(name: str, config: ClientConfiguration = default_configuration) -> Shelf:
    """
    Reads a Shelf from the Kubernetes API.

    :param name: The name of the Shelf to read.
    :param config: The configuration to use.
    :return: The Shelf.
    :raises RuntimeError: If the Shelf does not exist.
    """
    try:
        shelf = config.K8S_CUSTOM_OBJECT_API.get_namespaced_custom_object(
            group="beiboots.getdeck.dev",
            version="v1",
            namespace=config.NAMESPACE,
            plural="shelves",
            name=name,
        )
    except k8s.client.ApiException as e:  # type: ignore
        if e.status == 404:
            raise RuntimeError(f"The Shelf {name} does not exist")
        raise RuntimeError(str(e)) from None
    return Shelf(shelf)  # type: ignore



import base64
import logging
import socket
from pathlib import Path
from typing import List, Optional, Container

from beiboot.configuration import ClientConfiguration, __VERSION__

from beiboot.types import BeibootRequest, ShelfRequest

logger = logging.getLogger(__name__)


def create_beiboot_custom_ressource(
    req: BeibootRequest, config: ClientConfiguration
) -> dict:
    _labels = req.labels
    _labels.update({"beiboot.getdeck.dev/client-version": __VERSION__})
    cr = {
        "apiVersion": "getdeck.dev/v1",
        "kind": "beiboot",
        "provider": "k3s",
        "parameters": req.parameters.as_dict(),
        "fromShelf": req.from_shelf,
        "metadata": {
            "name": req.name,
            "namespace": config.NAMESPACE,
            "labels": _labels,
        },
    }
    return cr


def create_shelf_custom_ressource(
    req: ShelfRequest, config: ClientConfiguration
) -> dict:
    _labels = req.labels
    _labels.update({"beiboot.getdeck.dev/client-version": __VERSION__})
    cr = {
        "apiVersion": "beiboots.getdeck.dev/v1",
        "kind": "shelf",
        "clusterName": req.cluster_name,
        "volumeSnapshotClass": req.volume_snapshot_class,
        "volumeSnapshotContents": req.volume_snapshot_contents,
        "metadata": {
            "name": req.name,
            "namespace": config.NAMESPACE,
            "labels": _labels,
        },
    }
    return cr


def decode_kubeconfig(kubeconfig_obj: dict):
    """
    It takes a dictionary with a key called `source` and returns the value of that key decoded from base64

    :param kubeconfig_obj: The kubeconfig object that we created in the previous step
    :type kubeconfig_obj: dict
    :return: A string of the kubeconfig
    """
    b64_kubeconfig = kubeconfig_obj["source"]
    kubeconfig = base64.b64decode(b64_kubeconfig.encode("utf-8")).decode("utf-8")
    return kubeconfig


def get_beiboot_config_location(config: ClientConfiguration, cluster_name: str) -> str:
    """
    It creates a directory for the cluster's configuration file if it doesn't already exist, and returns the path to
    that directory

    :param config: ClientConfiguration
    :type config: ClientConfiguration
    :param cluster_name: The name of the cluster you want to create
    :type cluster_name: str
    :return: The path to the directory where the kubeconfig file will be stored.
    """
    config_dir = config.KUBECONFIG_LOCATION.joinpath(cluster_name)
    config_dir.mkdir(parents=True, exist_ok=True)
    return str(config_dir)


def get_kubeconfig_location(config: ClientConfiguration, cluster_name: str) -> str:
    """
    It returns the location of the kubeconfig file for a given cluster name

    :param config: This is the configuration object that we created earlier
    :type config: ClientConfiguration
    :param cluster_name: The name of the Beiboot cluster you want to create
    :type cluster_name: str
    :return: The path to the kubeconfig file for the cluster.
    """
    return str(
        Path(get_beiboot_config_location(config, cluster_name)).joinpath(
            f"{cluster_name}.yaml"
        )
    )


def decode_b64_dict(b64_dict: dict[str, str]) -> dict[str, str]:
    """
    It takes a dictionary of base64 encoded strings and returns a dictionary of decoded strings

    :param b64_dict: a dictionary of key-value pairs where the value is a base64-encoded string
    :type b64_dict: dict[str, str]
    :return: A dictionary with the keys and values decoded from base64.
    """
    return {
        k: base64.b64decode(v.encode("utf-8")).decode("utf-8").strip()
        for k, v in b64_dict.items()
    }


def _check_port_free(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("127.0.0.1", port))
        except OSError:
            raise RuntimeError(
                f"Can not establish requested connection: localhost:{port} is busy."
            ) from None
        else:
            s.close()
    return True


def _list_containers_by_prefix(
    config: ClientConfiguration, prefix: str
) -> List[Optional[Container]]:
    """
    It returns a list of all containers whose name starts with a given prefix

    :param config: ClientConfiguration
    :type config: ClientConfiguration
    :param prefix: The prefix of the container name
    :type prefix: str
    :return: A list of containers that start with the prefix.
    """

    containers = config.DOCKER.containers.list()
    result = []
    for container in containers:
        if container.name.startswith(prefix):  # type: ignore
            result.append(container)
    return result



import kubernetes as k8s

from beiboot.configuration import ClientConfiguration
from beiboot import api


def remove_all_beiboots(config: ClientConfiguration):
    bbts = api.read_all(config=config)
    for bbt in bbts:
        api.delete(bbt)


def remove_remainder_beiboot_namespaces(config: ClientConfiguration):
    try:
        configmap = config.K8S_CORE_API.read_namespaced_config_map(
            name=config.CONFIGMAP_NAME, namespace=config.NAMESPACE
        )
    except k8s.client.exceptions.ApiException:  # type: ignore
        return []
    if "namespacePrefix" not in configmap.data:  # type: ignore
        return []
    namespace_prefix = configmap.data["namespacePrefix"]  # type: ignore
    namespaces = config.K8S_CORE_API.list_namespace()
    removed_namespaces = []
    for ns in namespaces.items:
        if ns.metadata.name.startswith(namespace_prefix):
            removed_namespaces.append(ns.metadata.name)
    return removed_namespaces


def remove_remainder_bbts(config: ClientConfiguration):
    try:
        bbts = config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object(
            group="getdeck.dev",
            version="v1",
            namespace=config.NAMESPACE,
            plural="beiboots",
        )
    except Exception:
        return None
    for bbt in bbts.get("items"):
        try:
            config.K8S_CUSTOM_OBJECT_API.patch_namespaced_custom_object(
                group="getdeck.dev",
                version="v1",
                plural="beiboots",
                namespace=config.NAMESPACE,
                name=bbt["metadata"]["name"],
                body={"metadata": {"finalizers": None}},
            )
            config.K8S_CUSTOM_OBJECT_API.delete_namespaced_custom_object(
                group="getdeck.dev",
                version="v1",
                plural="beiboots",
                namespace=config.NAMESPACE,
                name=bbt["metadata"]["name"],
            )
        except Exception:
            continue
    return None


def remove_beiboot_crds(config: ClientConfiguration):
    try:
        config.K8S_EXTENSIONS_API.delete_custom_resource_definition(
            name="beiboots.getdeck.dev"
        )
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            return
        else:
            raise e from None
    try:
        config.K8S_EXTENSIONS_API.delete_custom_resource_definition(
            name="shelves.beiboots.getdeck.dev"
        )
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            return
        else:
            raise e from None


def remove_beiboot_rbac(config: ClientConfiguration):
    try:
        config.K8S_RBAC_API.delete_cluster_role_binding(name="getdeck-beiboot-operator")
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            pass
        else:
            raise e from None
    try:
        config.K8S_RBAC_API.delete_cluster_role(name="getdeck:beiboot:operator")
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            pass
        else:
            raise e from None


def remove_beiboot_webhooks(config: ClientConfiguration):
    try:
        config.K8S_ADMISSION_API.delete_validating_webhook_configuration(
            name="beiboot.getdeck.dev"
        )
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            pass
        else:
            raise e from None


def remove_beiboot_namespace(config: ClientConfiguration):
    try:
        config.K8S_CORE_API.delete_namespace(name=config.NAMESPACE)
    except k8s.client.exceptions.ApiException as e:  # type: ignore
        if e.status == 404:
            return
        else:
            raise e from None



import logging
from typing import Optional, List

from beiboot.api.utils import stopwatch
from beiboot.configuration import default_configuration, ClientConfiguration
from beiboot.connection.abstract import AbstractConnector
from beiboot.connection.factory import connector_factory
from beiboot.connection.types import ConnectorType
from beiboot.types import Beiboot

logger = logging.getLogger(__name__)


def _get_connector(
    connector_type: ConnectorType, config: ClientConfiguration
) -> AbstractConnector:
    connector = connector_factory.get(
        connector_type=connector_type,
        configuration=config,
    )
    return connector


@stopwatch
def connect(
    beiboot: Beiboot,
    connector_type: ConnectorType,
    host: Optional[str] = None,
    config: ClientConfiguration = default_configuration,
    _docker_network: Optional[str] = None,
) -> AbstractConnector:
    """
    Connects to a Beiboot instance.

    :param beiboot: The Beiboot instance to connect to.
    :type beiboot: Beiboot
    :param connector_type: The connector type to use.
    :type connector_type: ConnectorType
    :param host: The host to connect to.
    :param config: The client configuration to use.
    :type config: ClientConfiguration

    :return: A AbstractConnector instance
    """
    additional_ports: List = []
    connector = _get_connector(connector_type, config)

    if connector_type == ConnectorType.GHOSTUNNEL_DOCKER and _docker_network:
        # this is for local testing purposes
        connector.set_docker_network(_docker_network)  # type: ignore

    connector.establish(beiboot, additional_ports, host)
    return connector


@stopwatch
def terminate(
    name: str,  # passing only the name, as the Beiboot may already deleted
    connector_type: ConnectorType,
    config: ClientConfiguration = default_configuration,
) -> AbstractConnector:
    """
    Terminate a Beiboot connection and clean up

    :param name: The Beiboot instance name
    :type name: str
    :param connector_type: The connector type to use.
    :type connector_type: ConnectorType

    :return: A AbstractConnector instance
    """
    connector = _get_connector(connector_type, config)
    connector.terminate(name=name)
    connector.delete_beiboot_config_directory(beiboot_name=name)
    return connector



# flake8: noqa
from beiboot.types import InstallOptions


def data(params: InstallOptions) -> list[dict]:
    return [
        {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "metadata": {
                "name": "beiboot-operator",
                "namespace": params.namespace,
                "labels": {"app": "beiboot-operator"},
            },
            "spec": {
                "replicas": 1,
                "selector": {"matchLabels": {"app": "beiboot-operator"}},
                "template": {
                    "metadata": {"labels": {"app": "beiboot-operator"}},
                    "spec": {
                        "serviceAccountName": "beiboot-operator",
                        "containers": [
                            {
                                "name": "beiboot",
                                "image": f"quay.io/getdeck/beiboot:{params.version}",
                                "imagePullPolicy": "Always",
                                "ports": [{"containerPort": 9443}],
                            }
                        ],
                    },
                },
            },
        }
    ]
